/*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 */

/* $Id: linux_c.c 242 2006-08-20 19:24:15Z gerd $ */

/* Missing calls to Linux libc functions */

/* Linux: make all system prototypes available */
#define _GNU_SOURCE

#include "caml/mlvalues.h"
#include "caml/alloc.h"
#include "caml/fail.h"
#include "caml/memory.h"
#include <memory.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
/* #include <sys/capability.h> */
#include <net/if.h>
#include <sys/ioctl.h>
#include <linux/ioctl.h>
#include <linux/if_tun.h>
#include <sys/types.h>
#include <grp.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>

/**********************************************************************/
/* From unixsupport.h                                                 */
/**********************************************************************/

#define Nothing ((value) 0)

extern void unix_error (int errcode, char * cmdname, value arg) Noreturn;
extern void uerror (char * cmdname, value arg) Noreturn;

/**********************************************************************/

value unix_initgroups(value user, value gid) {
    char *user_s;
    long gid_i;
    int e;

    user_s = String_val(user);
    gid_i = Long_val(gid);

    e = initgroups(user_s, gid_i);
    if (e < 0) uerror("initgroups", Nothing);

    return Val_unit;
}


/* There are a number of bugs in Linux' getgrouplist implementation:
 * - Memory corruption for ngroups = 0:
 *   http://sources.redhat.com/bugzilla/show_bug.cgi?id=661
 * - Buffer overrun if user is member of too many groups
 *   https://bugzilla.redhat.com/bugzilla/show_bug.cgi?id=101691
 *   http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=245029
 *   http://www.securitytracker.com/alerts/2003/Oct/1007940.html
 *   (CAN-2003-0689)
 *
 * getgrouplist is not part of LSB 2.0.
 *
 * Because of these reasons, we do not use it for now, but implement
 * it on top of getgrent.
 */

#if 0
value unix_getgrouplist(value user, value gid) {
    char *user_s;
    long gid_i;
    gid_t *groups;
    int e, m, k;
    value r;

    user_s = String_val(user);
    gid_i = Long_val(gid);

    /* As described in getgrouplist(3), the function is broken in many
     * versions of glibc, esp. 2.3.2, because the boundary of the group
     * array is ignored. Workaround: Determine a good value first.
     */
    m = sysconf(_SC_NGROUPS_MAX);
    if (m < 0 || m > 500) m = 500;

    groups = (gid_t *) malloc( m * sizeof(gid_t) );
    if (groups == NULL) 
	failwith("getgrouplist: Cannot malloc - Out of memory");

    e = getgrouplist(user_s, gid_i, groups, &m);
    if (e < 0) {
	free(groups);
	failwith("getgrouplist: Too many supplementary groups");
    }

    r = alloc_tuple(m);
    /* Note: user and user_s are now invalid, i.e. may have been collected */

    for (k = 0; k < m; k++) {
	Field(r, k) = Val_long(groups[k]);
    };

    free(groups);

    return r;
}
#endif

value unix_getgrouplist(value user, value gid) {
    char *user_s;
    char *copy_s;
    long gid_i;
    struct group *g;
    char **member;
    int is_member;
    int n, k;
    value r;

    user_s = String_val(user);
    gid_i = Long_val(gid);

    /* Make a copy of user_s, because user_s will become invalid after
     * the first allocation on the O'Caml heap:
     */
    copy_s = stat_alloc(strlen(user_s) + 1);
    strcpy(copy_s, user_s);

    /* First loop: Count the number of matching groups */
    n = 0;

    /* Rewind group cursor to beginning of group list: */
    setgrent();

    /* Get first group record: */
    g = getgrent();
    while (g != NULL) {
	/* Is user_s member of this group? */
	is_member = 0;
	member = g->gr_mem;
	while (!is_member && *member != NULL) {
	    is_member = (strcmp(copy_s, *member) == 0);
	    member++;
	}

	if (is_member && g->gr_gid != gid_i) n++;

	/* Get next group record: */
	g = getgrent();
    }

    /* Close group file */
    endgrent();

    r = alloc_tuple(n+1);
    /* Note: user and user_s are now invalid, i.e. may have been collected */

    /* The main group is always the first element of the result array.
     * If it is also a supplementary group, it is not returned twice.
     */
    Field(r, 0) = Val_long(gid_i);

    /* Second loop: put matching groups into r */
    k = 0;

    /* Rewind group cursor to beginning of group list: */
    setgrent();

    /* Get first group record: */
    g = getgrent();
    while (g != NULL) {
	/* Is user_s member of this group? */
	is_member = 0;
	member = g->gr_mem;
	while (!is_member && *member != NULL) {
	    is_member = (strcmp(copy_s, *member) == 0);
	    member++;
	}

	if (is_member && g->gr_gid != gid_i) {
	    if (k < n) Field(r, k+1) = Val_long(g->gr_gid);
	    k++;
	}

	/* Get next group record: */
	g = getgrent();
    }

    /* Close group file */
    endgrent();

    /* Assert n = k, otherwise memory will get corrupted and pgm will segfault
     */
    if (n != k) {
	for (k=0; k <= n; k++) Field(r, k) = Val_long(0);
	failwith("getgrouplist: Number of groups changed while determining the list");
    }

    free(copy_s);

    return r;
}


/**********************************************************************/

#if 0
value unix_drop_privileges (value u) {
    /* ML = Cap.drop_privileges */
    /* Remove the following capabilities immediately (in principle, all but
     * CAP_SETUID and CAP_SETGID):
     * 
     * - CAP_CHOWN
     * - CAP_DAC_OVERRIDE
     * - CAP_DAC_READ_SEARCH
     * - CAP_FOWNER
     * - CAP_KILL
     * - CAP_SETPCAP
     * - CAP_LINUX_IMMUTABLE
     * - CAP_NET_BIND_SERVICE
     * - CAP_NET_BROADCAST
     * - CAP_NET_ADMIN
     * - CAP_NET_RAW
     * - CAP_IPC_LOCK
     * - CAP_IPC_OWNER
     * - CAP_SYS_MODULE
     * - CAP_SYS_RAWIO
     * - CAP_SYS_CHROOT
     * - CAP_SYS_PTRACE
     * - CAP_SYS_PACCT
     * - CAP_SYS_ADMIN
     * - CAP_SYS_BOOT
     * - CAP_SYS_NICE
     * - CAP_SYS_RESOURCE
     * - CAP_SYS_TIME
     * - CAP_SYS_TTY_CONFIG
     * - CAP_MKNOD
     * - CAP_LEASE
     *
     * Remove only in the inherited set (we need them to become a non-root user):
     * - CAP_SETUID
     * - CAP_SETGID
     */
    cap_t cap;
    cap_value_t caps_needed[2] = { CAP_SETUID, CAP_SETGID };

    cap = cap_get_proc();
    cap_clear(cap);
    cap_set_flag(cap, CAP_PERMITTED, 2, caps_needed, CAP_SET);
    cap_set_flag(cap, CAP_EFFECTIVE, 2, caps_needed, CAP_SET);
    if (cap_set_proc(cap) < 0) {
	uerror("cap_set_proc", Nothing);
    }
    return Val_unit;
}
#endif

/**********************************************************************/

value misc_sysconf_pagesize (value unit) {
    /* ML = Misc.sysconf_pagesize */
    return Val_long(sysconf(_SC_PAGESIZE));
}

/**********************************************************************/

value unix_grantpt (value fd) {
    /* ML = Pty.grantpt */
    int e;
    e = grantpt(Int_val(fd));
    if ( e < 0 ) uerror("grantpt", Nothing);
    return Val_unit;
}


value unix_unlockpt (value fd) {
    /* ML = Pty.unlockpt */
    int e;
    e = unlockpt(Int_val(fd));
    if ( e < 0 ) uerror("unlockpt", Nothing);
    return Val_unit;
}


value unix_ptsname (value fd) {
    /* ML = Pty.ptsname */
    char *s;

    s = ptsname(Int_val(fd));
    if ( s == NULL ) uerror("ptsname", Nothing);
    return copy_string(s);
}

/**********************************************************************/

value unix_realpath (value name) {
    /* ML = File.realpath */
    char *name_in_s, *name_out_s;
    value name_out;

    name_in_s = String_val(name);
    name_out_s = realpath(name_in_s, NULL);   /* Note: GNU extension! */
    if (name_out_s == NULL) {
	uerror("realpath", Nothing);
    }
    else {
	name_out = copy_string(name_out_s);
	free(name_out_s);
    }
    return name_out;
}


value unix_mknod (value name, value perm, value nt) {
    mode_t m;
    dev_t d;
    int e;

    m = Long_val(perm) & 07777;
    if (Is_block(nt)) {
	switch (Tag_val(nt)) {
	case 0:  /* = S_IFCHR */
	    m |= S_IFCHR;
	    d = Long_val(Field(nt,0));
	    break;
	case 1:  /* = S_IFBLK */
	    m |= S_IFBLK;
	    d = Long_val(Field(nt,0));
	    break;
	}
    }
    else {
	switch (Long_val(nt)) {
	case 0:  /* = S_IFREG */
	    m |= S_IFREG; break;
	case 1:  /* = S_IFIFO */
	    m |= S_IFIFO; break;
	case 2:  /* = S_IFSOCK */
	    m |= S_IFSOCK; break;
	}

    }

    e = mknod(String_val(name), m, d);
    if (e < 0) uerror("mknod", Nothing);

    return Val_unit;
}


/**********************************************************************/

value ml_set_sched_other ( value prio ) {
    struct sched_param p;
    int e;

    p.sched_priority = 0;
    e = sched_setscheduler(0, SCHED_OTHER, &p);
    if (e < 0) uerror("sched_setscheduler", Nothing);

    if (Is_block(prio)) {
	errno = 0;
	setpriority(PRIO_PROCESS, 0, Int_val(Field(prio,0)));
	if (errno < 0) uerror("setpriority", Nothing);
    };

    return Val_unit;
}


value ml_set_sched_rr ( value prio ) {
    struct sched_param p;
    int e;

    p.sched_priority = Long_val(prio);
    e = sched_setscheduler(0, SCHED_RR, &p);
    if (e < 0) uerror("sched_setscheduler", Nothing);

    return Val_unit;
}


/*
value ml_set_affinity ( value cpus ) {
    cpu_set_t cpu_set;
    CPU_ZERO(&cpu_set);
    e = sched_setaffinity(0, &cpu_set);
    -- There are 2-param versions of sched_setaffinity (e.g. Ubuntu Hoary)
    -- and 3-param versions (e.g. Debian Sarge).
    ...
}
*/


/**********************************************************************/


value tun_create_tap (value fd_val, value device_name_val, value device_owner_val) {
    /* ML = Tun.tun_create_tap */
    int fd;
    char *device_name;
    int device_owner;
    struct ifreq ifr;

    fd = Int_val(fd_val);
    device_name = String_val(device_name_val);
    device_owner = Int_val(device_owner_val);

    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    strncpy(ifr.ifr_name, device_name, sizeof(ifr.ifr_name) - 1);
    if(ioctl(fd, TUNSETIFF, (void *) &ifr) < 0){
	uerror("tun_create", Nothing);
    }
    if(ioctl(fd, TUNSETPERSIST, 1) < 0){
	uerror("tun_create", Nothing);
    }
    if(ioctl(fd, TUNSETOWNER, device_owner) < 0){
	uerror("tun_create", Nothing);
    }
    return copy_string(ifr.ifr_name);
}


value tun_delete_tap (value fd_val, value device_name_val) {
    /* ML = Tun.tun_delete_tap */
    int fd;
    char *device_name;
    struct ifreq ifr;

    fd = Int_val(fd_val);
    device_name = String_val(device_name_val);

    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    strncpy(ifr.ifr_name, device_name, sizeof(ifr.ifr_name) - 1);
    if(ioctl(fd, TUNSETIFF, (void *) &ifr) < 0){
	uerror("tun_delete", Nothing);
    }
    if(ioctl(fd, TUNSETPERSIST, 0) < 0){
	uerror("tun_delete", Nothing);
    }

    return Val_unit;
}
