(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: mount.ml 187 2005-10-14 10:54:15Z gerd $ *)

exception Already_mounted
exception Not_a_mount_point
exception Mount_point_not_empty
exception Mount_error

let ws_re = Str.regexp "[ \t\r\n]+";;

let read_proc_mounts() =
  let f = open_in "/proc/mounts" in
  let l = ref [] in
  try
    while true do
      let line = input_line f in
      let fields = Str.split ws_re line in
      match fields with
	| _ :: mpoint :: _ ->
	    l := mpoint :: !l
	| _ ->
	    ()
    done;
    assert false
  with
    | End_of_file ->
	close_in f;
	List.rev !l
    | error ->
	close_in f;
	raise error
;;


let is_mount_point link =
  (* Does [link] exist? *)
  if Sys.file_exists link then (
    (* Make [link] absolute: *)
    let link_abs =
      if link <> "" && link.[0] = '/' then
	link
      else
	Filename.concat (Sys.getcwd()) link in
    (* Make [link] the physical path: *)
    let link_phys = Linux.File.realpath link_abs in
    (* Read /proc/mounts: *)
    let mounts = read_proc_mounts() in
    List.mem link_phys mounts
  )
  else
    false
;;


let mount_bind_dir target link =
  if is_mount_point link then
    raise Already_mounted;

  if Sys.file_exists link then (
    (* Check that [link] is an empty directory: *)
    let s = Unix.LargeFile.stat link in
    match s.Unix.LargeFile.st_kind with
      | Unix.S_DIR ->
	  if Sys.readdir link <> [| |] then
	    raise Mount_point_not_empty
      | _ ->
	  raise Mount_point_not_empty
  ) 
  else (
    (* Create [link]: *)
    Unix.mkdir link 0o755;
  );

  Unix.chown link 0 0;    (* root becomes owner *)
  Unix.chmod link 0o755;  (* set privileges *)

  (* Finally, try the mount: *)
  let code =
    Sys.command (Printf.sprintf
                   "mount -o bind %s %s"
                   (Filename.quote target) 
		   (Filename.quote link)) in
  if code <> 0 then
    raise Mount_error
;;


let mount_bind_file target link =
  if is_mount_point link then
    raise Already_mounted;

  if Sys.file_exists link then (
    (* Check that [link] is an empty file: *)
    let s = Unix.LargeFile.stat link in
    match s.Unix.LargeFile.st_kind with
      | Unix.S_REG ->
	  if s.Unix.LargeFile.st_size <> 0L then
	    raise Mount_point_not_empty
      | _ ->
	  raise Mount_point_not_empty
  ) 
  else (
    (* Create [link]: *)
    let f = open_out link in
    close_out f;
  );

  Unix.chown link 0 0;    (* root becomes owner *)
  Unix.chmod link 0;      (* set privileges *)

  (* Finally, try the mount: *)
  let code =
    Sys.command (Printf.sprintf
                   "mount -o bind %s %s"
                   (Filename.quote target) 
		   (Filename.quote link)) in
  if code <> 0 then
    raise Mount_error
;;


let mount_tmpfs size link =
  if is_mount_point link then
    raise Already_mounted;

  if Sys.file_exists link then (
    (* Check that [link] is an empty directory: *)
    let s = Unix.LargeFile.stat link in
    match s.Unix.LargeFile.st_kind with
      | Unix.S_DIR ->
	  if Sys.readdir link <> [| |] then
	    raise Mount_point_not_empty
      | _ ->
	  raise Mount_point_not_empty
  ) 
  else (
    (* Create [link]: *)
    Unix.mkdir link 0;
  );

  Unix.chown link 0 0;    (* root becomes owner *)
  Unix.chmod link 0;      (* set privileges *)

  let code =
    Sys.command (Printf.sprintf
                   "mount -t tmpfs -o size=%dM,mode=700,noatime,nosuid,nodev none %s"
                   size
		   (Filename.quote link)) in
  if code <> 0 then
    raise Mount_error
;;


let umount link =
  if not(is_mount_point link) then
    raise Not_a_mount_point;

  let code =
    Sys.command (Printf.sprintf
                   "umount %s"
		   (Filename.quote link)) in
  if code <> 0 then
    raise Mount_error
;;
