(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umladmin.ml 201 2005-10-17 11:43:53Z gerd $ *)

open Umlmonctrl_aux
open Umlmon_types
open Umlmon_common

open Printf
open Cryptokit

module Mon = Umlmonctrl_clnt.Ctrl.V1

module StrMap = Map.Make(String)


let is_tty() =
  try
    ignore(Unix.tcgetattr Unix.stdin);
    true
  with
      _ -> false
;;


let tty_setup ?(interrupt=false) ?(crnl=false) f =
  let attr = Unix.tcgetattr Unix.stdin in
  let attr' = 
    { attr with
	Unix.c_ignbrk = not interrupt; (* maybe no CTRL-C = interrupt *)
	c_inlcr = false;      (* No LF -> CR mapping on input *)
	c_icrnl = crnl;       (* Maybe no CR -> LF mapping on input *)
	c_igncr = false;      (* Keep CR *)
	c_istrip = false;     (* 8 bit input *)
	c_ixon = false;       (* No XON/XOFF for input *)
	c_ixoff = false;      (* No XON/XOFF for input *)

	c_opost = false;      (* No output processing *)

	c_isig = interrupt;   (* No keyboard signals *)
	c_icanon = false;     (* No canonical mode *)
	c_echo = false;       (* Disable echo *)
	c_echoe = false;
	c_echok = false;
	c_echonl = false;

	c_vmin = 1;
	c_vtime = 0;
    } in
  Unix.tcsetattr Unix.stdin Unix.TCSANOW attr';
  try
    let r = f() in
    Unix.tcsetattr Unix.stdin Unix.TCSANOW attr;
    r
  with
      err ->
	Unix.tcsetattr Unix.stdin Unix.TCSANOW attr;
	raise err
;;


let call_proc id f arg =
  let ues = Unixqueue.create_unix_event_system() in
  let client = Umlmon_client.create_client ues id in
  try
     let r = f client arg in
     Rpc_client.shut_down client;
     r
  with
    | err ->
        Rpc_client.shut_down client;
        raise err
;;


type status_or_off = [ uml_status | `off ]


let status_map vm_list : (status_or_off StrMap.t) =
  List.fold_left
    (fun map vm ->
       let status =
	 try
	   let ues = Unixqueue.create_unix_event_system() in
	   let client = Umlmon_client.create_client ues vm in
	   try
	     ( Mon.get_status client () :> status_or_off )
	   with
	       any -> Rpc_client.shut_down client; raise any
	 with
	     any -> `off in
       StrMap.add vm status map
    )
    StrMap.empty
    vm_list
;;


let config_map vm_list : (umlspec StrMap.t) =
  List.fold_left
    (fun map vm ->
       try
	 let config = Umlmon_config.read_vm_config vm in
	 StrMap.add vm config map
       with
	 | Failure _ -> map
    )
    StrMap.empty
    vm_list
;;


let cmd_list list_mode args =
  let vm_base_list =
    if args = [] then 
      Umlmon_config.vm_list()
    else
      args in

  let map = status_map vm_base_list in
  StrMap.iter
    (fun vm status ->
       let status_str =
	 match status with
	   | `ctrl_inactive -> "INACTIVE"
	   | `ctrl_starting -> "STARTING"
	   | `ctrl_stopped -> "STOPPED"
	   | `ctrl_running -> "RUNNING"
	   | `off -> "OFF" in
       let selected =
	 match list_mode with
	   | `All      -> true
	   | `Active   -> status <> `ctrl_inactive && status <> `off
	   | `Inactive -> status = `ctrl_inactive
	   | `Off      -> status = `off in
       if selected then (
	 if list_mode = `All then
	   printf "%-20s %s\n" vm status_str
	 else
	   printf "%s\n" vm
       )
    )
    map;
  flush stdout
;;


let start_monitor vm =
  printf "  Monitoring %s\n" vm;
  flush stdout;
  ignore(Sys.command(sprintf "umlmon %s" (Filename.quote vm)))
;;


let stop_monitor vm =
  printf "  Shutting down monitor for %s\n" vm;
  flush stdout;
  let pid_file = Filename.concat "/var/run/umlmon" (vm ^ ".pid") in
  try
    if Sys.file_exists pid_file then (
      let f = open_in pid_file in
      let pid = input_line f in
      close_in f;
      Unix.kill (int_of_string pid) Sys.sigterm;
    )
    else (
      printf "  No PID file found!\n";
      flush stdout
    )
  with
      _ -> printf "  Error\n"; flush stdout
;;


let boot_vm vm =
  printf "  Booting %s\n" vm;
  flush stdout;
  try
    match call_proc vm Mon.start [| |] with
      | `code_error msg ->
	  printf "  Error: %s\n" msg
      | _ -> ()
  with
      _ -> printf "  Error\n"; flush stdout
;;


let shutdown_vm vm =
  printf "  Shutting down %s\n" vm;
  flush stdout;
  try
    match call_proc vm Mon.ctrlaltdel () with
      | `code_error msg ->
	  printf "  Error: %s\n" msg
      | _ -> ()
  with
      _ -> printf "  Error\n"; flush stdout
;;


let terminate_vm vm =
  printf "  Terminating %s\n" vm;
  flush stdout;
  try
    List.iter
      (fun f ->
	 match call_proc vm f () with
	   | `code_error msg ->
	       printf "  Error: %s\n" msg
	   | _ -> ()
      )
      [ (fun client () -> Mon.sysrq client "s");
        (fun client () -> Mon.sysrq client "u");
        Mon.halt ]
  with
      _ -> printf "  Error\n"; flush stdout
;;


let cmd_start args =
  let vm_list =
    if args = [] then 
      Umlmon_config.vm_list()
    else
      args in

  let s_map = status_map vm_list in
  let c_map = config_map vm_list in

  StrMap.iter
    (fun vm spec ->
       let status = StrMap.find vm s_map in
       match spec.umlstart with
	 | `Off ->
	     (* Leaving off! *)
	     ()
	 | `Manual ->
	     (* Only ensure that the monitor is started: *)
	     if status = `off then
	       start_monitor vm;
	 | `Boot ->
	     (* Additionally boot the VM: *)
	     ( match status with
		 | `off ->
		     start_monitor vm;
		     boot_vm vm
		 | `ctrl_inactive ->
		     boot_vm vm
		 | _ ->
		     ()
	     )
    )
    c_map
;;


let perform_stop timeout to_shutdown to_stop =
  (* First phase: Send CTRL-ALT-DEL *)
  
  List.iter shutdown_vm to_shutdown;

  (* Wait... *)

  let t0 = Unix.gettimeofday() in
  
  ( try
      while true do
	Unix.sleep 1;
	if Unix.gettimeofday() > t0 +. timeout then raise Exit;
	let s_map' = status_map to_shutdown in
	if StrMap.fold (fun vm status p -> 
			  match status with
			    | `off -> p
			    | `ctrl_inactive -> p
			    | _ -> false) s_map' true 
	then
	  raise Exit
      done
    with
      | Exit -> ()
  );

  (* Now stop monitors: *)

  let s_map' = status_map to_stop in
  List.iter
    (fun vm ->
       let status = StrMap.find vm s_map' in
       match status with
	 | `off -> ()
	 | `ctrl_inactive ->
	     stop_monitor vm
	 | _ ->
	     terminate_vm vm;
	     stop_monitor vm
    )
    to_stop
;;


let cmd_stop timeout args =
  let vm_list =
    if args = [] then 
      Umlmon_config.vm_list()
    else
      args in

  let s_map = status_map vm_list in

  let to_shutdown, to_stop =
    StrMap.fold
      (fun vm status (l_shutdown, l_stop) ->
	 match status with
	   | `off ->
	       (l_shutdown, l_stop)
	   | `ctrl_inactive ->
	       (l_shutdown, vm :: l_stop)
	   | _ ->
	       (vm :: l_shutdown, vm :: l_stop)
      )
      s_map
      ([], []) in

  perform_stop timeout to_shutdown to_stop
;;



let cmd_update timeout args =
  let vm_list =
    if args = [] then 
      Umlmon_config.vm_list()
    else
      args in

  let s_map = status_map vm_list in
  let c_map = config_map vm_list in

  let to_shutdown, to_stop, to_start =
    StrMap.fold
      (fun vm config (l_shutdown, l_stop, l_start) ->
	 let status = StrMap.find vm s_map in
	 let do_shutdown, do_stop, do_start =
	   match config.umlstart with
	     | `Off ->
		 (status <> `off && status <> `ctrl_inactive,
		  status <> `off,
		  false)
	     | `Manual
	     | `Boot ->
		 (false,
		  false,
		  status = `off) in
	 (if do_shutdown then vm :: l_shutdown else l_shutdown),
	 (if do_stop then vm :: l_stop else l_stop),
	 (if do_start then vm :: l_start else l_start)
      )
      c_map
      ([], [], []) in

  perform_stop timeout to_shutdown to_stop;
  List.iter start_monitor to_start
;;


let cmd_password() =
  let pw =
    if is_tty() then (
      prerr_string "Password: ";
      flush stderr;
      let pw = tty_setup ~interrupt:true ~crnl:true read_line in
      prerr_newline();
      flush stderr;
      pw
    )
    else (
      read_line()
    ) in

  let encoder = Hexa.encode() in
  print_endline (transform_string encoder (Digest.string pw));
  flush stdout
;;


let cmd_config clear_list set_list rename_opt vm =
  let config = Umlmon_config.read_vm_config vm in
  let ini = Umlmon_config.get_inifile ~priv:true vm config in
  List.iter
    (fun name ->
       let n = List.length (ini # getaval vm name) in
       for k = 1 to n do 
	 ini # delval vm name
       done
    )
    clear_list;
  List.iter
    (fun (name,value) ->
       ini # setval vm name value
    )
    set_list;
  let ini' =
    match rename_opt with
      | None -> 
	  ini
      | Some r ->
	  let ini' = new Inifiles.inifile "/dev/null" in
	  ini # iter (fun n v -> ini' # setval r n v) vm;
	  ini'
  in
  ini' # save ~file:"/dev/stdout" ()
;;


let read_lines file =
  let rec read_rest f =
    try
      let line = input_line f in
      line :: read_rest f
    with
	End_of_file -> []
  in
  let f = open_in file in
  try
    let lines = read_rest f in
    close_in f;
    lines
  with
      err -> close_in f; raise err
;;


let write_lines lines file =
  let f = open_out file in
  try
    List.iter (fun line -> output_string f line; output_string f "\n") lines;
    close_out f
  with
      err -> close_out f; raise err
;;


let nv_re = Str.regexp "^\\([a-zA-Z0-9_]+\\)=\\(.*\\)$";;

let version() =
  printf "UMLMON Release %s Revision %s\n" Instconf.release Instconf.revision;
  printf "Copyright 2005 by Informatikbuero Gerd Stolpmann.\n\n";
  printf "UMLMON is free software, and can be distributed under the terms\n";
  printf "of the GNU General Public License.\n\n";
  printf "Commercial support available, please contact gerd@gerd-stolpmann.de\n";
  flush stdout;
  exit 0
;;

let main() =
  let list_mode = ref `All in
  let cmd = ref None in
  let args = ref [] in
  let timeout = ref 300.0 in
  let clear_list = ref [] in
  let set_list = ref [] in
  let rename = ref None in

  Arg.parse
    (Arg.align
       [ "-active", Arg.Unit (fun () -> list_mode := `Active),
	 " List only active VMs";

	"-inactive", Arg.Unit (fun () -> list_mode := `Inactive),
	 " List only inactive but monitored VMs";

	"-off", Arg.Unit (fun () -> list_mode := `Off),
	 " List only VMs not monitored";

	"-timeout", Arg.Set_float timeout,
	"<n> Wait this number of seconds for VM shutdown";

	"-clear", Arg.String (fun s -> clear_list := s :: !clear_list),
	"<param> config: delete this parameter in the printed configuration";

	"-set", Arg.String (fun s -> 
			      if Str.string_match nv_re s 0 then
				set_list := (Str.matched_group 1 s,
					     Str.matched_group 2 s) :: !set_list
			      else
				raise(Arg.Bad ("Bad syntax: " ^ s))),
	"<param>=<value> config: add this parameter in the printed configuration";

	"-rename", Arg.String (fun s -> rename := Some s),
	"<id> config: rename the VM in the printed configuration";

	"-version", Arg.Unit version, " Print version number and exit";
       ])
    (fun s ->
       match !cmd with
	 | None -> cmd := Some s
	 | Some _ -> args := !args @ [s] )
    "usage: umladmin [options] (list|start|stop|update|password|config) [ vm ... ]

list:      List available VM monitors
start:     Start VM monitor, and if configured, also boot the VM
stop:      Shutdown the VM and stop the VM monitor
update:    Start and stop monitors as configured
password:  Encrypt password for inclusion into config file
config:    Extract the configuration of a VM

By default, all configured VM monitors are managed at once (except 'config'). 
You can give a list of VM monitors on the command line. Only local monitors 
can be managed!
";

  match !cmd with
    | None ->
	failwith "No command given"
    | Some "list" ->
	cmd_list !list_mode !args
    | Some "start" ->
	cmd_start !args
    | Some "stop" ->
	cmd_stop !timeout !args
    | Some "update" ->
	cmd_update !timeout !args
    | Some "password" ->
	cmd_password ()
    | Some "config" ->
	( match !args with
	    | [vm] -> cmd_config (List.rev !clear_list) (List.rev !set_list) !rename vm
	    | _ -> failwith "Please pass the name of exactly one VM to 'config'!"
	)
    | Some s ->
	failwith ("Unknown command: " ^ s)
;;


try
  Sys.set_signal Sys.sigpipe Sys.Signal_ignore;
  Unix.putenv "PATH" (Filename.concat Instconf.prefix "sbin" ^
		   ":/bin:/sbin:/usr/bin:/usr/sbin");
  ignore(Unix.umask 0o22);
  main();
with
  | Failure msg ->
      prerr_endline ("umladmin: " ^ msg); exit 1
  | Sys_error msg ->
      prerr_endline ("umlmdmin: " ^ msg); exit 1
  | Unix.Unix_error(_,_,_) as e ->
      prerr_endline ("umladminn: " ^ Umlmon_common.string_of_unix_error e); exit 1
  | err ->
      prerr_endline ("umladmin: Caught exception " ^ Printexc.to_string err); exit 1
;;
