(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_common.ml 187 2005-10-14 10:54:15Z gerd $ *)

open Umlmon_types
open Printf
open Scanf

let rec has_dups l =
  match l with
    | [] -> false
    | x :: l' -> 
	List.mem x l' || has_dups l'


let rec restart f arg =
  try
    f arg
  with
      Unix.Unix_error(Unix.EINTR,_,_) ->
	restart f arg


let last_msg = ref None
let last_msg_t = ref 0.0
let last_msg_rep = ref 0
let last_msg_sup = ref 0  (* how often suppressed *)

let monitor_log msg =
  let t = Unix.gettimeofday() in
  let tm = Unix.gmtime(Unix.gettimeofday()) in
  let tm_str = Printf.sprintf "[%04d-%02d-%02d %02d:%02d:%02d UTC]"
		 (tm.Unix.tm_year + 1900)
		 (tm.Unix.tm_mon + 1)
		 tm.Unix.tm_mday
		 tm.Unix.tm_hour
		 tm.Unix.tm_min
		 tm.Unix.tm_sec in
  let do_print = ref true in
  if !last_msg = Some msg then (
    incr last_msg_rep;
    if !last_msg_rep > 100 then (
      if t -. !last_msg_t > 60.0 && !last_msg_sup > 0 then (
	Printf.eprintf "%s (Last message was repeated %d times in the last %.1f seconds)\n"
		tm_str !last_msg_sup (t -. !last_msg_t);
	last_msg_t := t;
	last_msg_sup := 0;
      )
      else
	incr last_msg_sup;
      do_print := false;
    )
  ) else (
    if !last_msg_sup >= 1 then (
      Printf.eprintf "%s (Last message was repeated %d times in the last %.1f seconds)\n"
				  tm_str !last_msg_sup (t -. !last_msg_t);
    );
    last_msg_t := t;
    last_msg_rep := 0;
    last_msg := Some msg;
    last_msg_sup := 0;
  );

  if !do_print then (
    Printf.eprintf "%s %s\n" tm_str msg
  );
  flush stderr
;;


let fd_log ?(nl = false) fd msg =
  let t = Unix.gmtime(Unix.gettimeofday()) in
  let s =
    Printf.sprintf 
      "%s[%04d-%02d-%02d %02d:%02d:%02d UTC] %s\n"
      (if nl then "\n" else "")
      (t.Unix.tm_year + 1900)
      (t.Unix.tm_mon + 1)
      t.Unix.tm_mday
      t.Unix.tm_hour
      t.Unix.tm_min
      t.Unix.tm_sec
      msg in
  ignore(Unix.write fd s 0 (String.length s));  (* assumes fd is a file *)
;;


let array_mem x a =
  try
    for k = 0 to Array.length a - 1 do
      if a.(k) = x then raise Exit
    done;
    false
  with Exit -> true
;;


let user_privileges user group_opt =
  let pw =
    try Unix.getpwnam user
    with Not_found -> failwith ("User does not exist: " ^ user) in
  let uid = pw.Unix.pw_uid in
  let gid_default = pw.Unix.pw_gid in
  let groups = Linux.Privilege.getgrouplist user gid_default in
  let gid =
    match group_opt with
      | None -> gid_default
      | Some group ->
	  let gr =
	    try Unix.getgrnam group
	    with Not_found -> failwith ("Group does not exist: " ^ group) in
	  let gid = gr.Unix.gr_gid in
	  if not (array_mem gid groups) then
	    failwith("User is not member of this group: " ^ group);
	  gid in
  { priv_uid = uid;
    priv_gid = gid;
    priv_groups = groups;
  }
;;
	  

let file_access filename p =
  let merge (r1,w1,x1) (r2,w2,x2) =
    (r1 || r2, w1 || w2, x1 || x2) in

  let access_of_bits n =
    (n land 4 > 0, n land 2 > 0, n land 1 > 0) in

  let s = Unix.LargeFile.stat filename in

  let owner_access =
    if p.priv_uid = s.Unix.LargeFile.st_uid then
      access_of_bits (s.Unix.LargeFile.st_perm lsr 6)
    else
      (false, false, false) in

  let group_access =
    if p.priv_gid = s.Unix.LargeFile.st_gid || 
       array_mem s.Unix.LargeFile.st_gid p.priv_groups
    then
      access_of_bits (s.Unix.LargeFile.st_perm lsr 3)
    else
      (false, false, false) in

  let world_access =
    access_of_bits s.Unix.LargeFile.st_perm in

  merge (merge owner_access group_access) world_access
;;


let open_log_file umlspec name =
  try
    Unix.openfile name [ Unix.O_WRONLY; Unix.O_APPEND ] 0 
  with
      Unix.Unix_error(Unix.ENOENT,_,_) ->
	let fd = Unix.openfile name [ Unix.O_WRONLY; Unix.O_APPEND; Unix.O_CREAT] 0o666 in
	Unix.fchown fd umlspec.umlpriv.priv_uid umlspec.umlpriv.priv_gid;
	fd

let reopen_stderr spec =
  match spec.umlmonlog with
    | None -> ()
    | Some name ->
	let fd = open_log_file spec name in
	Unix.dup2 fd Unix.stderr;
	Unix.close fd



(* Note for [protect]: Some procedures are not encapsulated by [protect]. Exceptions
 * must be caught nevertheless, but this must happen in the procedures themselves.
 * List:
 * - read
 * - write
 *)

let protect proc session arg reply =
  try
    try
      proc session arg reply
    with
      | Sys.Break as exn -> 
	  raise exn
      | Auth_failed _ as err ->
	  Rpc_server.reply_error session Rpc.Auth_failed;
      | Rpc_server.Connection_lost ->
	  (* Just ignore this. *)
	  ()
      | err ->
	  Rpc_server.reply_error session Rpc.System_err;
	  monitor_log ("umlmon: Caught exception: " ^ Printexc.to_string err)
  with
    | Rpc_server.Connection_lost ->
	()


let rpc_over_inet session =
  match Rpc_server.get_socket_name session with
    | Unix.ADDR_UNIX _ -> false
    | Unix.ADDR_INET _ -> true

let daemon ?(wait=false) pidfile f =
  (* Run f() as a daemon, without controlling terminal. Note that all terminals
   * must be opened with O_NOCTTY to prevent them from becoming controlling terminals!
   *
   * [wait]: if true, wait until the daemon terminates.
   *)
  match Unix.fork() with
    | 0 ->
	( match Unix.fork() with
	    | 0 ->
		ignore(Unix.setsid());
		Unix.chdir "/";
		Unix.close Unix.stdin;
		ignore(Unix.openfile "/dev/null" [Unix.O_RDONLY] 0);
		Unix.close Unix.stdout;
		ignore(Unix.openfile "/dev/null" [Unix.O_WRONLY] 0);
		let pf = open_out pidfile in
		output_string pf (string_of_int (Unix.getpid()));
		output_string pf "\n";
		close_out pf;
		Sys.set_signal Sys.sigterm (Sys.Signal_handle
					      (fun _ -> raise Sys.Break));
		( try
		    f();
		  with
		      Sys.Break -> ()
		);
		( try Sys.remove pidfile with _ -> ());
		exit 0
	    | n ->
		Sys.set_signal Sys.sigint (Sys.Signal_handle
					     (fun _ -> Unix.kill n Sys.sigint));
		Sys.set_signal Sys.sigterm (Sys.Signal_handle
					     (fun _ -> Unix.kill n Sys.sigterm));
		if wait then ignore(restart (Unix.waitpid []) n);
		Unix_exts._exit 0   (* don't run atexit *)
	)
    | n ->
	Sys.set_signal Sys.sigint (Sys.Signal_handle
				     (fun _ -> Unix.kill n Sys.sigint));
	Sys.set_signal Sys.sigterm (Sys.Signal_handle
				      (fun _ -> Unix.kill n Sys.sigterm));
	if wait then ignore(restart (Unix.waitpid []) n);
	Unix_exts._exit 0   (* don't run atexit *)
;;


let string_of_unix_error exn =
  match exn with
    | Unix.Unix_error(e,_,arg) ->
	let m = Unix.error_message e in
	if arg = "" then m else m ^ ": " ^ arg
    | _ ->
	assert false
;;


let format_hwaddr (MAC s) =
  sprintf 
    "%02x:%02x:%02x:%02x:%02x:%02x"
    (Char.code s.[0])
    (Char.code s.[1])
    (Char.code s.[2])
    (Char.code s.[3])
    (Char.code s.[4])
    (Char.code s.[5])
;;


let parse_hwaddr s =
  try
    let (c0,c1,c2,c3,c4,c5) =
      sscanf
	s
	"%2x:%2x:%2x:%2x:%2x:%2x%!" 
	(fun c0 c1 c2 c3 c4 c5 -> (c0,c1,c2,c3,c4,c5))
    in
    let s' = String.make 6 ' ' in
    s'.[0] <- Char.chr c0;
    s'.[1] <- Char.chr c1;
    s'.[2] <- Char.chr c2;
    s'.[3] <- Char.chr c3;
    s'.[4] <- Char.chr c4;
    s'.[5] <- Char.chr c5;
    MAC s'
  with
    | Scan_failure _ ->
	failwith("Invalid MAC address: " ^ s)
;;


let increment_hwaddr (MAC s) =
  let s = String.copy s in
  let rec increment n =
    let c = Char.code(s.[n]) in
    let c' = if c = 255 then 0 else c+1 in
    s.[n] <- Char.chr c';
    if c' = 0 && n > 0 then increment(n-1)
  in
  increment 5;
  MAC s
;;
