(*
 * <COPYRIGHT>
 * Copyright 2005 Gerd Stolpmann
 *
 * <GPL>
 * This file is part of UMLMON.
 *
 * UMLMON is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * UMLMON is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with UMLMON; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * </>
 *)

(* $Id: umlmon_mconsole.ml 187 2005-10-14 10:54:15Z gerd $ *)

open Umlmon_types
open Umlmon_common


exception Timeout

(* Old implementation of mconsole: *)

(*
let mconsole gs args =       (* may raise Shell.Subprocess_error *)
  let umlspec = Lazy.force gs.umlspec in
  let mconsole_socket =
    Printf.sprintf "%s/tmp/uml/%s/mconsole"
      umlspec.umljaildir
      umlspec.umlid
  in
  let b = Buffer.create 256 in
  let ues = Unixqueue.create_unix_event_system() in
  let eng =
    new Shell_uq.call_engine
      ~stdin:Shell.from_dev_null
      ~stdout:(Shell.to_buffer b)
      [ Shell.command 
	  ~arguments:(Array.append [| mconsole_socket |] args)
	  "uml_mconsole" ]
      ues in
  let eng' =
    new Uq_engines.watchdog 5.0 eng in   (* mconsole must respond within 5 seconds *)
  Uq_engines.when_state
    ~is_error:(fun _ -> eng # abort())
    eng';
  Uq_engines.when_state
    ~is_done:(fun _ -> eng' # abort())
    ~is_error:(fun _ -> eng' # abort())
    eng;
  Unixqueue.run ues;
  match eng # state with
    | `Done _ ->
	Buffer.contents b
    | `Error err ->
	raise err
    | `Aborted ->
	raise Timeout
    | _ ->
	assert false
;;
 *)


type mconsole_result =
    [ `OK of string
    | `Error of string
    ]


let rec bind_client s prefix n =
  if n > 100 then failwith "Too many client sockets";
  try
    let name = prefix ^ string_of_int n in
    Unix.bind s (Unix.ADDR_UNIX name);
    name
  with
      Unix.Unix_error(Unix.EADDRINUSE,_,_) ->
	bind_client s prefix (n+1)
;;


let unlink_client =
  function
    | None -> ()
    | Some n -> Unix.unlink n
;;


let mconsole gs args =
  let umlspec = Lazy.force gs.umlspec in
  let mconsole_socket =
    Printf.sprintf "%s/tmp/uml/%s/mconsole"
      umlspec.umljaildir
      umlspec.umlid
  in
  let s = Unix.socket Unix.PF_UNIX Unix.SOCK_DGRAM 0 in
  let s_name = ref None in
  try 
    Unix.set_nonblock s;
    Unix.connect s (Unix.ADDR_UNIX mconsole_socket);
    (* s_name := Some(bind_client s "/tmp/.mconsoleclient" 0); *)
    Unix.bind s (Unix.ADDR_UNIX ("\000" ^ string_of_int (Unix.getpid())));
    (* Don't know why - only this strange address works *)

    let cmd = String.concat " " args in
    let req_block = Mconsole.mk_request cmd in
    let (_,outl,_) = restart(Unix.select [] [s] []) 5.0 in  (* not fully correct... *)
    if outl = [] then raise Timeout;
    ignore(Unix.send s req_block 0 (String.length req_block) []);

    let r = ref [] in
    let cont = ref true in
    let buf = String.create 1024 in  (* large enough *)
    while !cont do
      let (inl, _, _) = restart(Unix.select [s] [] []) 5.0 in  (* not fully correct... *)
      if inl = [] then raise Timeout;
      let n = Unix.recv s buf 0 (String.length buf) [] in
      let reply_block = String.sub buf 0 n in
      let (err, more, data) = Mconsole.dest_reply reply_block in
      if err then
	r := (`Error data) :: !r
      else
	r := (`OK data) :: !r;
      cont := (not err) && more;
    done;

    Unix.close s;
    unlink_client !s_name;
    List.rev !r
  with
    | Unix.Unix_error(_,_,_) as e ->
	Unix.close s;
	unlink_client !s_name;
	let m = Umlmon_common.string_of_unix_error e in
	monitor_log ("umlmon: While contacting MCONSOLE: " ^ m);
	failwith("While contacting MCONSOLE: " ^ m)
    | err ->
	Unix.close s;
	unlink_client !s_name;
	raise err
;;


let proc_mconsole_query gs f =
  protect 
    (fun session arg reply ->
       Umlmon_auth.ensure_authenticated session;
       try
	 let uml_answer = mconsole gs (f arg) in
	 match uml_answer with
	     [ `OK s ] -> reply (Some s)
	   | _ -> reply None
       with
	 | Failure _ ->
	     reply None
	 | Timeout ->
	     monitor_log ("umlmon: Timeout while connecting to MCONSOLE socket");
	     reply None
    )
;;

let proc_get_ssl gs =
  proc_mconsole_query gs (fun n -> [ "config"; "ssl" ^ Int32.to_string n ]) ;;
  
let proc_get_con gs =
  proc_mconsole_query gs (fun n -> [ "config"; "con" ^ Int32.to_string n ]) ;;

let proc_get_ubd gs =
  proc_mconsole_query gs (fun n -> [ "config"; "ubd" ^ Int32.to_string n ]) ;;

let proc_get_eth gs =
  proc_mconsole_query gs (fun n -> [ "config"; "eth" ^ Int32.to_string n ]) ;;


let proc_mconsole_signal gs name f transition =
  protect 
    (fun session arg reply ->
       Umlmon_auth.ensure_authenticated session;
       try
	 monitor_log ("umlmon: Signalling '" ^ name ^ "' to VM");
	 let new_vm = transition gs.umlvm in
	 let uml_answer = mconsole gs (f arg) in
	 match uml_answer with
	   | [ `OK _ ] -> 
	       gs.umlvm <- new_vm;
	       reply `code_ok
	   | [`Error msg] -> 
	       reply (`code_error ("Error from MCONSOLE: " ^ msg))
	   | _ -> 
	       reply (`code_error "Unexpected answer from MCONSOLE")
       with
	 | Timeout ->
	     monitor_log ("umlmon: Timeout while connecting to MCONSOLE socket");
	     reply (`code_error "MCONSOLE timeout")
	 | Failure msg ->
	     monitor_log ("umlmon: " ^ msg);
	     reply (`code_error msg)
    )
;;



let proc_stop gs =
  proc_mconsole_signal 
    gs
    "STOP"
    (fun () -> ["stop"]) 
    (function
       | `Inactive ->  
	   failwith "Instance is inactive"
       | `Finishing ->  
	   failwith "Instance is currently shutting down"
       | `Running actv ->
	   `Stopped actv
       | `Stopped actv ->
	   `Stopped actv
       | `Starting _ ->
	   failwith "Instance not running"
    )
;;


let proc_continue gs =
  proc_mconsole_signal 
    gs
    "GO"
    (fun () -> ["go"]) 
    (function
       | `Inactive ->  
	   failwith "Instance is inactive"
       | `Finishing ->  
	   failwith "Instance is currently shutting down"
       | `Running actv ->
	   `Running actv
       | `Stopped actv ->
	   `Running actv
       | `Starting _ ->
	   failwith "Instance not running"
    )
;;


let proc_halt gs =
  proc_mconsole_signal 
    gs
    "HALT"
    (fun () -> ["halt"]) 
    (function
       | `Inactive ->  
	   failwith "Instance is inactive"
       | `Finishing ->  
	   failwith "Instance is currently shutting down"
       | `Running actv ->
	   `Running actv
       | `Stopped actv ->
	   `Running actv   (* implies GO *)
       | `Starting _ ->
	   failwith "Instance not running"
    )
;;


let proc_ctrlaltdel gs =
  proc_mconsole_signal 
    gs
    "CAD"
    (fun () -> ["cad"]) 
    (function
       | `Inactive ->  
	   failwith "Instance is inactive"
       | `Finishing ->  
	   failwith "Instance is currently shutting down"
       | `Running actv ->
	   `Running actv
       | `Stopped actv ->
	   `Running actv   (* implies GO *)
       | `Starting _ ->
	   failwith "Instance not running"
    )
;;


let proc_sysrq gs =
  proc_mconsole_signal 
    gs
    "SYSRQ"
    (fun letter -> ["sysrq"; letter]) 
    (function
       | `Inactive ->  
	   failwith "Instance is inactive"
       | `Finishing ->  
	   failwith "Instance is currently shutting down"
       | `Running actv ->
	   `Running actv
       | `Stopped actv ->
	   `Stopped actv
       | `Starting _ ->
	   failwith "Instance not running"
    )
;;
